/* SmartMedia nandflash block device driver.
 * Written by Nikita V. Youshchenko <yoush@cs.msu.su>.
 * 
 * Used information from the following sources:
 * 
 *   LDD: Writing Linux Device Drivers, 2nd edition
 *        available at http://www.oreilly.com/catalog/linuxdrive2/
 *        
 *   SS: official Samsung's reference code for SmartMedia cards, smfs221.zip,
 *       downoaded from http://darkfader.net/gp32/smc.html
 *       Note: I used this code only to learn how to control hardware, not a
 *       single line was copied.
 *
 *   MTD: linux MTD drivers, http://www.linux-mtd.infradead.org/
 *
 *   SM: official Samsung's datasheet about SmartMedia 64M and 128M cards
 *
 *   SSFDC: physical format specification from SSFDC forum;
 *
 *   And possibly some other sources that I don't remember ...
 */


#include <linux/malloc.h>	/* kmalloc() & co */
#include <linux/string.h>	/* memset() & co */
#include <linux/init.h>		/* (?) for putting code in init section */
#include <linux/genhd.h>	/* for partitioned device support */
#include <asm/hardware.h>	/* IO_PBDR & co */
#include <linux/delay.h>	/* udelay() */
#include <asm/uaccess.h>	/* access_ok(), put_user(), etc */
#include <linux/hdreg.h>	/* struct hd_geometry */
#include <asm/spinlock.h>	/* locking */
#include <linux/timer.h>	/* timer organization */
#include <linux/sm.h>		/* driver-specific stuff */

#include "nand_ecc.h"		/* this header and corresponding implementation is from MTD */

#include <linux/major.h>	/* SM_MAJOR defined there */
#ifndef SM_MAJOR
#error Seems that invalid linux/major.h was included
#endif

#define SM_SHIFT	4	/* Max 15 partitions per device (just for fun ...) */
#define SM_MAXDEVS	1	/* Currently only one device (/dev/sma) is supported */
#define SM_TOTALMINORS	(SM_MAXDEVS << SM_SHIFT)
#define SM_BLOCKSIZE	1024	/* Seems to be lowest supported value */
/* According to LDD, the following should be BEFORE including linux/blk.h */
#define MAJOR_NR	SM_MAJOR
#define DEVICE_NR(device) (MINOR(device) >> SM_SHIFT)	/* from LDD */
#define DEVICE_NAME	"sm"
#define DEVICE_REQUEST	sm_request
/* The following three are defined to empty strings */
#define DEVICE_NO_RANDOM
#define DEVICE_ON(dev)
#define DEVICE_OFF(dev)
/* At last we are ready to include it ... */
#include <linux/blk.h>

/* Function prototypes. Put it here to declare static objects at the top
 * of the file (I like this style) */

/* Forward declaration of device description structure */
struct sm_device;

/* Function that checks for a media change */
static int sm_check_media_change(kdev_t kdev);			/* wrapper for struct file_operations */
static int sm_do_check_media_change(struct sm_device *dev);	/* real operation; must be called with lock held */
/* Function that detects the media and builds the mapping table */
static int sm_revalidate(kdev_t kdev);				/* wrapper for struct file_operation */
static int sm_do_revalidate(struct sm_device *dev);		/* real operation; must be called with lock held */
/* Function called to open the device */
static int sm_open(struct inode *inode, struct file *file);
/* Function called to close ("release") the device */
static int sm_close(struct inode *inode, struct file *file);
/* ioctl() handler */
static int sm_ioctl(struct inode *inode, struct file *file, unsigned cmd, unsigned long arg);
/* fsync handler. Is needed to flush durty buffers (hmm... what to do when port to 2.4 will be needed... */
static int sm_fsync(struct file *file, struct dentry *dentry);
/* Function sm_request() that actually does the I/O is declared in linux/blk.h */

/* Device description for the block device layer.
 * In 2.4 kernels, struct block_device_operations is used for it.
 * For 2.2, struct file_operations is */
static struct file_operations sm_ops = {
	open:			sm_open,
	release:		sm_close,
	ioctl:			sm_ioctl,
	check_media_change:	sm_check_media_change,
	revalidate:		sm_revalidate,
	read:			block_read,
	write:			block_write,
	fsync:			sm_fsync,
};

/* Sizes of all devices (including patritions) controlled by this driver.
 * Needed for block device layer, for blk_size[MAJOR_NR].
 * sm_gendisk.sizes also points to this array. */
static int sm_devsizes[SM_TOTALMINORS];

/* Block sizes for all devices (including partitions) controlled by this driver.
 * Needed for block device layer, for blksize_size[MAJOR_NR].
 * All elements are initialized to SM_BLOCKSIZE in sm_init() */
static int sm_blocksizes[SM_TOTALMINORS];

/* Struct gendisk object is needed for partitions support.
 * LDD advices to set most fields statically. */
static struct hd_struct sm_part[SM_TOTALMINORS];
static void dummy_gendisk_init(struct gendisk *unused) {}
static struct gendisk sm_gendisk = {
	major:			MAJOR_NR,
	major_name:		DEVICE_NAME,
	minor_shift:		SM_SHIFT,
	max_p:			1 << SM_SHIFT,	/* Why not ...-1?  One minor number is for the whole device ...  But LDD says do this way. */
	max_nr:			SM_MAXDEVS,
	init:			dummy_gendisk_init,	/* stub needed: it is really called */
	part:			&(sm_part[0]),
	sizes:			&(sm_devsizes[0]),
	nr_real:		0 /*SM_MAXDEVS*/,	/* for now, don't want partition check on boot */
/* Other fields are dynamic */
};


/* Known card parameters table (for now, only cards with 512 byte pages are supported) */
static struct sm_params {
	unsigned char id;		/* Identification magic */
	unsigned short pagesinblock;	/* Number of pages in a block */
	unsigned short pblocks;		/* Number of physical blocks */
	unsigned short lblocks;		/* Number of logical blocks, according to SSFDC */
	unsigned short cylinders, heads, sectors; /* For pseudo-disk; according to SSFDC */
	const char *name;		/* Device name */
} known_cards[] = {
	{ 0x6B, 16, 512, 500, 250, 4, 8, "5V 4M card" },
	{ 0xD5, 16, 512, 500, 250, 4, 8, "3.3V 4M card" },
	{ 0xE3, 16, 512, 500, 250, 4, 8, "3.3V 4M card" },
	{ 0xE5, 16, 512, 500, 250, 4, 8, "3.3V 4M card" },
	{ 0xD6, 16, 1024, 1000, 250, 4, 16, "3.3V 8M card" },
	{ 0xE6, 16, 1024, 1000, 250, 4, 16, "3.3V 8M card" },
	{ 0x73, 32, 1024, 1000, 500, 4, 16, "3.3V 16M card" },
	{ 0x75, 32, 2048, 2000, 500, 8, 16, "3.3V 32M card" },
	{ 0x76, 32, 4096, 4000, 500, 8, 32, "3.3V 64M card" },
	{ 0x79, 32, 8192, 8000, 500, 16, 32, "3.3V 128M card" },
	{ 0, 0, 0, 0, 0 }
};

/* Known card vendor table */
static struct sm_vendor {
	unsigned char id;		/* Identification magic */
	const char *name;		/* Vendor name */
} known_vendors[] = {
	{ 0x98, "Toshiba" },
	{ 0xEC, "Samsung" },
	{ 0x04, "Fujitsu" },
	{ 0x8f, "National" },
	{ 0, "unknown" }
};

/* Device writes are hard because nobody knows the order of requests from block device layer,
 * but we can't write any page at any time. Seems that "block size" parameter of block device
 * is somewhat dummy, at least in 2.2.1 kernel - it shouts and burns if "block size" is not
 * equal to 1k.
 * We NEED to write complete device blocks, that is - 8 or 16 kilobytes.
 * Otherwise it becomes possible that on some moment different pages of the same lblock
 * are stored in different pblocks. Although we may keep a table that describes this, bad
 * times will happen on sync() time, and even worse - if card will be unexpectedly taken away.
 * Well maybe it is still possible to implement, but we will go another way.
 *
 * We will have buffers than will handle 'dirty' lblocks, one buffer per lblock.
 * Lblock is 'dirty' if it has pages that were requested to be written, but still are not.
 * On a write request, data is just buffered.
 * On buffer flush time, first any page not in buffer yet is loaded into the buffer.
 * Then a new pblock is allocated, and data is written there. At last, old pblock is erased.
 * On write errors complete block is marked as bad, and another pblock is allocated
 * in the same zone. If there are none, write is aborted, causing data loss. In this case
 * the old pblock will still contain old version of the lblock.
 * Buffers are flushed:
 *  - on sync time,
 *  - on unmount time,
 *  - when a new block is becoming 'dirty' and there is no free buffer,
 *  - every N seconds any 'dirty' block that was not modified last N seconds is flushed.
 */

struct dirty_buffer_page {
	unsigned char valid;		/* Non-zero if buffer does contain data */
	char data[SM_PAGE_SIZE];
	char spare[SM_SPARE_SIZE];
};
#define VALID_OLD 1	/* page data was read, ECC already known */
#define VALID_NEW 2	/* page data was written, ECC should be calculated */

struct dirty_buffer {
	unsigned long timestamp;		/* jiffies of the last modofocation */
	unsigned short lblock;			/* lblock stored in this buffer; UNUSED_BUFFER for free buffer */
	struct dirty_buffer_page pages[0];	/* real size is dev->params->pagesinblock */
};
#define UNUSED_BUFFER 0xFFFF

/* Driver will always try to allocate MAX_DIRTY_BUFFERS buffers.
 * Only kmalloc() fails may make it to use lower number of buffers */
#define MAX_DIRTY_BUFFERS 4

/* Period of buffer flushes, in seconds */
#define DIRTY_FLUSH_PERIOD 5

/* Device description structure */
struct sm_device {
	struct sm_params *params;	/* params for the current card */
	struct sm_vendor *vendor;	/* vendor of the current card */
	unsigned short *p2l;	/* Table to map physical block to logical block; zero if unread */
	unsigned short *l2p;	/* Table to map logical block to physical block; zero if unread */
	unsigned short nparts;		/* partition count */
	unsigned short nopen;		/* use count */
	unsigned short rawmode;		/* 1, if operations that may affect block mapping are in progress */
	unsigned short npartopen;	/* partition use count (for -EBUSY on part. table reread) */
	unsigned short nmapped, nfree, nbad, ncis;	/* Statistics: number of mapped, free, bad and cis blocks */
	unsigned short dirty_buffer_count;	/* total number of dirty buffers */
	struct dirty_buffer *dirty_buffers[MAX_DIRTY_BUFFERS];	/* dirty buffers */
	/* We use very dumb locking model: when any function runs that may either read or write
	 * anything from struct sm_dev, a spinlock is locked.
	 * Data consistency and hardware access becomes protected automatically. */
	spinlock_t lock;
	/* A timer is used for periodic buffer flushes */
	struct timer_list timer;
	/* In better 2.4 kernels there is del_timer_sync(), but we don't have it in 2.2.
	 * So we need an additional spinlock to avoid race between timer deletion and timer handler */
	spinlock_t timer_lock;
};
/* Device descriptions */
static struct sm_device sm_devices[SM_MAXDEVS];


/* Some constant data used in writes (initialized in sm_init() */
char allones_data[SM_PAGE_SIZE];	// A page buffer all filled with 0xFFs
char bad_spare[SM_SPARE_SIZE];		// Spare area for bad blocks


/* Macros to print device name in diagnostics */
#define DEVICE_NAME_F DEVICE_NAME "%c"
#define DEVICE_NAME_L(dev) ('a' + ((struct sm_device *)dev - &(sm_devices[0])))

/* Command codes */
#define SM_CMD_RESET		0xFF		/* Reset */
#define SM_CMD_READID		0x90		/* Read ID */
#define SM_CMD_READPAGE		0x00		/* Read page and possibly spare area*/
#define SM_CMD_READSPARE	0x50		/* Read spare area */
#define SM_CMD_READSTATUS	0x70		/* Read status register */
#define SM_CMD_PREPAREWRITE	0x80		/* Receive address and data to write page */
#define SM_CMD_DOWRITE		0x10		/* Write received data */
#define SM_CMD_PREPAREERASE	0x60		/* Receive address for erase */
#define SM_CMD_DOERASE		0xD0		/* Erase block */
/* Status register bits */
#define SM_STATUS_READY		0x40	/* Status register bit that indicates that device completed operation */
#define SM_STATUS_FAIL		0x01	/* Status register bit that indicates device failure */
/* Some delays (in microseconds) */
#define SM_READ_DELAY	12	/* Delay between read command and data availability */
#define SM_WRITE_DELAY	525	/* A little more than maximum page write time */
#define SM_ERASE_TIME	3250	/* A little more than maximum block erase time */
#define SM_RESET_DELAY	525	/* A little more than maximum device reset time */

/* All hardware access functions must be called with lock held */

/* Hardware-related constants related to port access */
#define SM_CE		0x40		/* Chip Enable (board-specific) */
#define SM_CLE		0x10		/* Command Latch Enable (board-specific) */
#define SM_ALE		0x20		/* Address Latch Enable (board-specific) */
/* Lowest level tools to control hardware */
static inline void sm_init_hardware(struct sm_device *dev)
{
	IO_PBDDR = SM_ALE | SM_CLE | SM_CE;
	IO_PBDR &= ~(SM_ALE | SM_CLE | SM_CE);
}
static inline void sm_chip_enable(struct sm_device *dev)
{
	IO_PBDR |= SM_CE;
}
static inline void sm_chip_disable(struct sm_device *dev)
{
	IO_PBDR &= ~SM_CE;
}
static inline void sm_command_latch_enable(struct sm_device *dev)
{
	IO_PBDR |= SM_CLE;
}
static inline void sm_command_latch_disable(struct sm_device *dev)
{
	IO_PBDR &= ~SM_CLE;
}
static inline void sm_address_latch_enable(struct sm_device *dev)
{
	IO_PBDR |= SM_ALE;
}
static inline void sm_address_latch_disable(struct sm_device *dev)
{
	IO_PBDR &= ~SM_ALE;
}
static inline unsigned char sm_inb(struct sm_device *dev)
{
	return IO_NANDF;
}
static inline void sm_outb(struct sm_device *dev, unsigned char data)
{
	IO_NANDF = data;
}

static inline void sm_put_command(struct sm_device *dev, unsigned char cmd)
{
	sm_command_latch_enable(dev);
	sm_outb(dev, cmd);
	sm_command_latch_disable(dev);
}
static inline void sm_put_address_id(struct sm_device *dev)
{
	sm_address_latch_enable(dev);
	sm_outb(dev, 0);
	sm_address_latch_disable(dev);
}
static inline void sm_put_address_rw(struct sm_device *dev, unsigned short pblock, unsigned short pageinblock)
{
	unsigned long page = (unsigned long)pblock * dev->params->pagesinblock + pageinblock;

	sm_address_latch_enable(dev);
	sm_outb(dev, 0);
	sm_outb(dev, page & 0xFF);
	sm_outb(dev, (page >> 8) & 0xFF);
	/* Code in SS uses 4 bytes for 64M and larger cards */
	if (dev->params->pblocks >= 4096)
		sm_outb(dev, (page >> 16) & 0x7F);
	sm_address_latch_disable(dev);
}
static inline void sm_put_address_er(struct sm_device *dev, unsigned short pblock)
{
	unsigned long page = (unsigned long)pblock * dev->params->pagesinblock;

	sm_address_latch_enable(dev);
	sm_outb(dev, page & 0xFF);
	sm_outb(dev, (page >> 8) & 0xFF);
	/* Code in SS uses 3 bytes for 64M and larger cards */
	if (dev->params->pblocks >= 4096)
		sm_outb(dev, (page >> 16) & 0x7F);
	sm_address_latch_disable(dev);
}

/* Function that performs device reset */
static void sm_reset(struct sm_device *dev)
{
	sm_chip_enable(dev);
	sm_put_command(dev, SM_CMD_RESET);
	udelay(SM_RESET_DELAY);	/* FIXME: a context switch may be better than udelay() here (but what about lock?) */
	sm_chip_disable(dev);
}

/* Function to read ID from hardware */
static unsigned short sm_read_id(struct sm_device *dev)
{
	unsigned char b1, b2;
	
	sm_chip_enable(dev);
	sm_put_command(dev, SM_CMD_READID);
	sm_put_address_id(dev);
	b1 = sm_inb(dev);
	b2 = sm_inb(dev);
	sm_chip_disable(dev);
	return (b1 << 8) | b2;
}

/* Function to read data and/or spare area of the given page */
static void sm_read_page(struct sm_device *dev, unsigned short pblock, unsigned short pageinblock, char *data, char *spare)
{
	int i;
	
	sm_chip_enable(dev);
	sm_put_command(dev, data ? SM_CMD_READPAGE : SM_CMD_READSPARE);
	sm_put_address_rw(dev, pblock, pageinblock);
	udelay(SM_READ_DELAY);
	if (data) {
		for (i = 0; i < SM_PAGE_SIZE; i++)
			data[i] = sm_inb(dev);
	}
	if (spare) {	/* spare area may be read just aftr data */
		for (i = 0; i < SM_SPARE_SIZE; i++)
			spare[i] = sm_inb(dev);
	}
	sm_chip_disable(dev);
	/* Device starts to prefetch the next page and does not respond to commands until it is done.
	 * To avoid bad things, just wait now */
	udelay(SM_READ_DELAY);
}

/* Function to read status register (chip should be enabled) */
static inline unsigned char sm_read_status(struct sm_device *dev)
{
	sm_put_command(dev, SM_CMD_READSTATUS);
	return sm_inb(dev);
}

/* Function to write data and spare area */
static int sm_write_page(struct sm_device *dev, unsigned short pblock, unsigned short pageinblock, char *data, char *spare)
{
	int i;
	unsigned char status;
	
	sm_chip_enable(dev);
	sm_put_command(dev, SM_CMD_READPAGE);	/* Reset page pointer, in case it points not to the page beginning. */
	sm_put_command(dev, SM_CMD_PREPAREWRITE);
	sm_put_address_rw(dev, pblock, pageinblock);
	for (i = 0; i < SM_PAGE_SIZE; i++)
		sm_outb(dev, data[i]);
	for (i = 0; i < SM_SPARE_SIZE; i++)
		sm_outb(dev, spare[i]);
	sm_put_command(dev, SM_CMD_DOWRITE);

	for (i = 0; i < SM_WRITE_DELAY / 5; i++ ) {
		udelay(5);
		status = sm_read_status(dev);
		if (status & SM_STATUS_READY)
			break;
	}
	status = sm_read_status(dev);
	sm_chip_disable(dev);
	if (!(status & SM_STATUS_READY)) {
		printk(KERN_ERR DEVICE_NAME_F ": write operation timeout, doing device reset\n", DEVICE_NAME_L(dev));
		sm_reset(dev);
		return -EIO;
	}
	if (status & SM_STATUS_FAIL)
		return -EIO;
	return 0;
}

/* Function to erase pblock */
static int sm_erase_block(struct sm_device *dev, unsigned short pblock)
{
	unsigned char status;
	
	sm_chip_enable(dev);
	sm_put_command(dev, SM_CMD_PREPAREERASE);
	sm_put_address_er(dev, pblock);
	sm_put_command(dev, SM_CMD_DOERASE);
	udelay(SM_ERASE_TIME);	/* FIXME: huge udelay()! Context switch will be better ... (but what about lock?) */
	status = sm_read_status(dev);
	sm_chip_disable(dev);
	if (status & SM_STATUS_FAIL)
		return -EIO;
	return 0;
}


/* For some strange reason, ECC calculated by MTD's nand_ecc.c has bytes 1 and 2 inverted.
 * Son't want to change their code, so workarounding this instead.
 * FIXME: find out why they do this */
static inline void sm_mirror_ecc(unsigned char *src, unsigned char *dst)
{
	dst[0] = src[1];
	dst[1] = src[0];
	dst[2] = src[2];
}
/* Calculate ECC */
static inline void sm_calculate_ecc(char *data, unsigned char *ecc)
{
	unsigned char tmp_ecc[3];
	nand_calculate_ecc(data, tmp_ecc);
	sm_mirror_ecc(tmp_ecc, ecc);
}

/* Calculate ECC and try to correct data.
 * Errors are reported to syslog; on error -EIO is returned.
 * FIXME: nothing is done for blocks with read errors */
static int sm_correct_data(struct sm_device *dev, unsigned short pblock, unsigned short pageinblock,
		unsigned short part, char *data, unsigned char *read_ecc)
{
	unsigned char tmp_read_ecc[3], tmp_calc_ecc[3];
	int res;
	sm_mirror_ecc(read_ecc, tmp_read_ecc);
	nand_calculate_ecc(data, tmp_calc_ecc);
	res = nand_correct_data(data, tmp_read_ecc, tmp_calc_ecc);
	sm_mirror_ecc(tmp_read_ecc, read_ecc);	/* In case of tmp_read_ecc correction ... */
	switch (res) {
		case 0:
			return 0;		/* No error */
		case 1:
			printk(KERN_NOTICE DEVICE_NAME_F ": corrected one-bit error in data for part %d of page %u of pblock %u)\n",
					DEVICE_NAME_L(dev), part, pageinblock, pblock);
			return 0;
		case 2:
			printk(KERN_NOTICE DEVICE_NAME_F ": corrected one-bit error in stored ECC for part %d of page %u of pblock %u)\n",
					DEVICE_NAME_L(dev), part, pageinblock, pblock);
			return 0;
		default:
			printk(KERN_ERR DEVICE_NAME_F ": unable to correct error in part %d of page %u of block %u\n",
					DEVICE_NAME_L(dev), part, pageinblock, pblock);
			return -EIO;
	}
}

/* Read physical page and correct it using ECC */
static int sm_read_and_correct_page(struct sm_device *dev, unsigned short pblock, unsigned short pageinblock,
		char *data, char *spare)
{
	int res1, res2; 
	/* FIXME: the following code is for 512 byte pages only */
	sm_read_page(dev, pblock, pageinblock, data, spare);

	/* Do ECC check and correction */
	res1 = sm_correct_data(dev, pblock, pageinblock, 1, data, sm_spare_ecc1(spare));
	res2 = sm_correct_data(dev, pblock, pageinblock, 2, data + 256, sm_spare_ecc2(spare));
	
	return (res1 || res2) ? -EIO : 0;
}


/* Mark bad block.
 * SS first erases block, and then writes FF's into data area and bad block mark info spare area.
 * Do the same */
static void mark_bad_pblock(struct sm_device *dev, unsigned short pblock)
{
	unsigned short pageinblock;
	
	sm_erase_block(dev, pblock);
	for (pageinblock = 0; pageinblock < dev->params->pagesinblock; pageinblock++)
		sm_write_page(dev, pblock, pageinblock, allones_data, bad_spare);
}


/* Find an empty pblock from lblock zone */
static unsigned short find_empty_pblock(struct sm_device *dev, unsigned short lblock)
{
	static unsigned short rnd = 0x5555;
	unsigned short i, p, start, end;

	/* A pseudo-random generator. FIXME: use something better */
	rnd ^= jiffies & 0xFFFF;
	rnd += 359;

	/* Get zone boundaries */
	start = LZONE(lblock) * PBLOCKSINZONE;
	end = start + PBLOCKSINZONE;
	if (end > dev->params->pblocks)
		end = dev->params->pblocks;

	/* Find unused block in the zone */
	i = p = start + (rnd % (end - start));
	do {
		if (dev->p2l[i] == FREE_PBLOCK)
			return i;
		i++;
		if (i == end)
			i = start;
	} while (i != p);

	return NO_PBLOCK;		/* No free block in the zone */
}


/* Write back information from dirty buffer. */
static int flush_dirty_buffer(struct sm_device *dev, struct dirty_buffer *db)
{
	unsigned short old_pblock, new_pblock, baddr;
	int i, res;

	/* Do nothing if buffer is empty */
	if (db->lblock == UNUSED_BUFFER)
		return 0;

	/* Get old pblock used for this lblock, if any */
	old_pblock = dev->l2p[db->lblock];

	/* Get block address data field */
	baddr = sm_encode_baddr(db->lblock);

	/* Prepare data for writing */
	for (i = 0; i < dev->params->pagesinblock; i++) {
		if (db->pages[i].valid == VALID_NEW) {
			/* For updated pages, recalculate ECC */
			sm_calculate_ecc(db->pages[i].data, sm_spare_ecc1(db->pages[i].spare));
			sm_calculate_ecc(db->pages[i].data + 256, sm_spare_ecc2(db->pages[i].spare));
		} else if (db->pages[i].valid == VALID_OLD) {
			/* Do nothing for blocks that were read but not modified */
		} else if (old_pblock != FREE_PBLOCK) {
			/* If lblock is not new, load pages that were not updated from the old pblock.
	 		 * Errors at this stage are bad - they mean that old data may be lost.
	 		 * FIXME: find a better way to handle this */
			sm_read_and_correct_page(dev, old_pblock, i, db->pages[i].data, db->pages[i].spare);
			db->pages[i].valid = VALID_OLD;
		} else {
			/* We still will have to write new page because of it's spare area.
			 * As data, allones_page will be used. Spare should be prepared. */
			memset(db->pages[i].spare + 6, 0xFF, SM_SPARE_SIZE - 6);
		}

		/* Prepare spare area */
		sm_build_good_block_spare_start(db->pages[i].spare);	/* first 6 bytes affected */
		sm_put_baddr_to_spare(baddr, db->pages[i].spare);
	}

	/* Now we are ready to write */
retry:
	/* Choose pblock */
	new_pblock = find_empty_pblock(dev, db->lblock);
	/* Fail if there are no free blocks in the zone */
	if (new_pblock == NO_PBLOCK) {
		printk(KERN_ERR DEVICE_NAME_F ": too many bad blocks, nowhere to write lblock %d\n",
				DEVICE_NAME_L(dev), db->lblock);
		return -EIO;
	}

	/* We have allocated a free pblock */
	dev->nfree--;

	/* Try to write */
	for (i = 0; i < dev->params->pagesinblock; i++) {
		/* Do write */
		res = sm_write_page(dev, new_pblock, i,
					db->pages[i].valid ? db->pages[i].data : allones_data,
					db->pages[i].spare);
		/* On error, mark whole pblock as bad and retry whth another pblock */
		if (res) {
			mark_bad_pblock(dev, new_pblock);
			dev->p2l[new_pblock] = BAD_PBLOCK;
			dev->nbad++;
			goto retry;
		}
	}

	/* We have written a new pblock */
	dev->nmapped++;

	/* Now erase old pblock and fix mapping tables */
	dev->l2p[db->lblock] = new_pblock;
	dev->p2l[new_pblock] = db->lblock;
	if (old_pblock != FREE_PBLOCK) {
		dev->nmapped--;
		if (sm_erase_block(dev, old_pblock) == 0) {
			dev->p2l[old_pblock] = FREE_PBLOCK;
			dev->nfree++;
		} else {
			mark_bad_pblock(dev, old_pblock);
			dev->p2l[old_pblock] = BAD_PBLOCK;
			dev->nbad++;
		}
	}

	return 0;
}


/* Flush dirty buffers. If all==1, all buffers are flushed. Otherwise, buffers that
 * have no recent writes are flushed */
static int flush_dirty_buffers(struct sm_device *dev, int all)
{
	int i;
	struct dirty_buffer *b;
	int res = 0, res2;

	for (i = 0; i < dev->dirty_buffer_count; i++) {
		b = dev->dirty_buffers[i];
		if (b->lblock != UNUSED_BUFFER &&
				(all || (int)(jiffies - b->timestamp) > HZ * DIRTY_FLUSH_PERIOD)) {
			res2 = flush_dirty_buffer(dev, b);
			if (!res && res2)
				res = -EIO;
			b->lblock = UNUSED_BUFFER;
		}
	}

	return res;
}


/* Timer handler. Flush buffers */
static void sm_timeout(unsigned long param)
{
	struct sm_device *dev = (struct sm_device *)param;
	unsigned long flags;

	spin_lock_irqsave(&dev->timer_lock, flags);

	/* If device is no longer open, do nothing */
	if (!dev->nopen) {
		spin_unlock_irqrestore(&dev->timer_lock, flags);
		return;
	}

	/* If device lock is currently held, we cannot do anything.
	 * In this case, just reprogram timer for a recent moment */
	if (spin_trylock(&dev->lock) != 0) {
		dev->timer.expires = jiffies + 1;
		add_timer(&dev->timer);
		spin_unlock_irqrestore(&dev->timer_lock, flags);
		return;
	}

	/* Reprogram timer for the next tick and enable irq's.
	 * Then do the flushes with irq's enabled, but dev->lock held */
	dev->timer.expires = jiffies + HZ * DIRTY_FLUSH_PERIOD;
	add_timer(&dev->timer);
	spin_unlock_irqrestore(&dev->timer_lock, flags);

	flush_dirty_buffers(dev, 0);
	spin_unlock(&dev->lock);
	return;
}


/* Find dirty buffer for the selected lblock.
 * If allocate is non-zero, dirty buffer will be allocated if not present */
static struct dirty_buffer *get_dirty_buffer(struct sm_device *dev, unsigned short lblock, int allocate)
{
	int i;
	struct dirty_buffer *b, *c;

	/* Try to find already allocated buffer */
	for (i = 0; i < dev->dirty_buffer_count; i++) {
		b = dev->dirty_buffers[i];
		if (b->lblock == lblock)
			return b;
	}

	if (!allocate)
		return 0;
	
	/* Find best candidate for allocation.
	 * Empty buffer is always the best.
	 * If no buffer is empty, one with oldest timestamp is the best */
	for (i = 0, c = 0; i < dev->dirty_buffer_count; i++) {
		b = dev->dirty_buffers[i];
		if (b->lblock == UNUSED_BUFFER) {
			c = b;
			break;
		}
		if (!c || ((int)(b->timestamp - c->timestamp)) > 0)
			c = b;
	}

	/* If chosen buffer is not unused, it should be flushed first */
	if (c->lblock != UNUSED_BUFFER) {
		if (flush_dirty_buffer(dev, c) != 0)
			return 0;
	}

	/* Set lblock for allocated buffer and mark all pages in the buffer as not valid yet */
	c->lblock = lblock;
	for (i = 0; i < dev->params->pagesinblock; i++)
		c->pages[i].valid = 0;

	return c;
}


/* Read logical page (and correct it using ECC) */
static int sm_read_lpage(struct sm_device *dev, unsigned long lpage, char *buf)
{
	unsigned short lblock, pblock, pageinblock;
	char spare[SM_SPARE_SIZE];
	struct dirty_buffer *db;
	int res;
	
	/* Check bounds */
	if (lpage >= dev->params->lblocks * dev->params->pagesinblock) {
		/*printk(KERN_ERR DEVICE_NAME_F ": attempt to read after end of device\n", DEVICE_NAME_L(dev));*/
		return -EIO;
	}

	/* Get lblock of the request */
	lblock = lpage / dev->params->pagesinblock;
	pageinblock = lpage % dev->params->pagesinblock;

	/* Maybe dirty buffer is allocated for lblock, and the page is there */
	db = get_dirty_buffer(dev, lblock, 0);
	if (db && db->pages[pageinblock].valid) {
		memcpy(buf, db->pages[pageinblock].data, SM_PAGE_SIZE);
		return 0;
	}

	/* Device read needed. Map lblock to pblock */
	pblock = dev->l2p[lblock];

	/* Check for unmapped lblock.
	 * Unmapped lblocks are OK to read, they should be filled by ones */
	if (pblock == NO_PBLOCK) {
		/*printk(KERN_WARNING DEVICE_NAME_F
			": attempt to read unmapped block, returning 0xFF\'s\n", DEVICE_NAME_L(dev));*/
		memset(buf, 0xFF, SM_PAGE_SIZE);
		return 0;
	}

	/* Read corresponding page from pblock */
	res = sm_read_and_correct_page(dev, pblock, pageinblock, buf, spare);
	if (!res)
		return res;

	/* If dirty buffer is allocated, copy the page there */
	if (db) {
		db->pages[pageinblock].valid = VALID_OLD;
		memcpy(db->pages[pageinblock].data, buf, SM_PAGE_SIZE);
		memcpy(db->pages[pageinblock].spare, spare, SM_SPARE_SIZE);
	}

	return 0;
}

/* Write logical page to dirty buffer */
static int sm_write_lpage(struct sm_device *dev, unsigned long lpage, char *buf)
{
	unsigned short lblock, pageinblock;
	struct dirty_buffer *db;

	/* Find out lblock and get dirty buffer */
	lblock = lpage / dev->params->pagesinblock;
	pageinblock = lpage % dev->params->pagesinblock;
	db = get_dirty_buffer(dev, lblock, 1);
	if (!db) {
		printk(KERN_ERR DEVICE_NAME_F ": writing page %lu: failed to allocate dirty buffer\n",
				DEVICE_NAME_L(dev), lpage);
		return -EIO;
	}

	/* Copy data to dirty buffer. Actual write will happen later. */
	db->pages[pageinblock].valid = VALID_NEW;
	memcpy(db->pages[pageinblock].data, buf, SM_PAGE_SIZE);
	db->timestamp = jiffies;	/* update timestamp */

	return 0;
}

static void DEVICE_REQUEST(void)
{
	kdev_t kdev;
	unsigned char devn;
	struct sm_device *dev;
	unsigned long lpage, start, afterend;
	char *buf;
	struct hd_struct *hds;
	char nmbuf[8];

	/* Request may be both for whole device and for partition.
	 * Anyway, sm_gendisk.part[minor] may be used to map request to device lpages */
	while (1) {
		INIT_REQUEST;
		
		kdev = CURRENT->rq_dev;
		devn = DEVICE_NR(kdev);
		dev = &(sm_devices[devn]);
		hds = sm_gendisk.part + MINOR(kdev);
		start = CURRENT->sector + hds->start_sect;
		afterend = start + CURRENT->current_nr_sectors;
		spin_lock(&dev->lock);
		if (dev->rawmode) {
			printk(KERN_ERR "%s: request when device is in raw mode\n",
					disk_name(&sm_gendisk, MINOR(kdev), nmbuf));
			end_request(0);
			spin_unlock(&dev->lock);
			continue;
		}
		if (afterend > hds->start_sect + hds->nr_sects) {
			/*printk(KERN_ERR "%s: invalid request (jumps over end of device)\n",
					disk_name(&sm_gendisk, MINOR(kdev), nmbuf));*/
			end_request(0);
			spin_unlock(&dev->lock);
			continue;
		}

		for (lpage = start, buf = CURRENT->buffer; lpage < afterend;
				lpage++, buf += SM_PAGE_SIZE) {
			if (CURRENT->cmd == READ) {
				if (sm_read_lpage(dev, lpage, buf) < 0)
					break;
			} else if (CURRENT->cmd == WRITE) {
				if (sm_write_lpage(dev, lpage, buf) < 0)
					break;
			}
		}
		
		end_request(lpage == afterend);
		spin_unlock(&dev->lock);
	}

	spin_unlock(&dev->lock);
}


int __init sm_init(void)
{
	int i, res;
	struct sm_device *dev;
	
	/* Register block device */
	res = register_blkdev(MAJOR_NR, DEVICE_NAME, &sm_ops);
	if (res) {
		printk(KERN_ERR DEVICE_NAME ": unable to register block device at major %d\n",
				MAJOR_NR);
		return res;
	}

	/* Set request function */
	blk_dev[MAJOR_NR].request_fn = DEVICE_REQUEST;
	/* blk_size[MAJOR_NR] will be inited on partition check called from revaliate() */
	/* blksize_size[MAJOR_NR] needs manual initialization */
	for (i = 0; i < SM_TOTALMINORS; i++)
		sm_blocksizes[i] = SM_BLOCKSIZE;
	blksize_size[MAJOR_NR] = sm_blocksizes;
	/* For the rest block device parameter arrays defaults are used. */

	/* Connect our struct gendisk to system list */
	sm_gendisk.next = gendisk_head;
	gendisk_head = &sm_gendisk;
	
	/* Init hardware and sm_devices array */
	for (i = 0; i < SM_MAXDEVS; i++) {
		dev = &(sm_devices[i]);	/* static kernel data is pre-zeroed */
		sm_init_hardware(dev);
		sm_reset(dev);		/* just for case */
		dev->lock = SPIN_LOCK_UNLOCKED;
		init_timer(&dev->timer);
		dev->timer_lock = SPIN_LOCK_UNLOCKED;
	}
	
	/* Init constant data */
	memset(allones_data, 0xFF, SM_PAGE_SIZE);
	sm_build_bad_block_spare(bad_spare);
	
	printk("SmartMedia nandflash block device driver registered with major %d\n", MAJOR_NR);
	return 0;
}


static int sm_check_media_change(kdev_t kdev)
{
	struct sm_device *dev = &(sm_devices[DEVICE_NR(kdev)]);
	int res;
	
	/* With current dumb sm_do_check_media_change() lock may be not needed here.
	 * But when real check will be made, lock may become really required */
	spin_lock(&dev->lock);
	res = sm_do_check_media_change(dev);
	spin_unlock(&dev->lock);
	return res;
}

static int sm_do_check_media_change(struct sm_device *dev)
{
	/* FIXME: I don't know how to detect media change.
	 *        For now, I assume no media changes and return 1 only on the first call. */
	return dev->p2l == 0;
}



/* (Re-)Read partition table from the device. Shold be called with lock held. */
static int sm_read_part_table(struct sm_device *dev)
{
	unsigned short part1, npart, i;
	unsigned char devn = dev - sm_devices, devminor = devn << SM_SHIFT;

	part1 = devminor + 1;
	npart = (1 << SM_SHIFT) - 1;
	/* Set size of the whole device. Needed for resetup_one_dev() */
	sm_devsizes[devminor] = dev->params->lblocks * dev->params->pagesinblock / (SM_BLOCKSIZE / SM_PAGE_SIZE);
	sm_gendisk.part[devminor].start_sect = 0;
	sm_gendisk.part[devminor].nr_sects = dev->params->lblocks * dev->params->pagesinblock;
	/* Clear buffer that will be filled by resetup_one_dev() */
	memset(sm_gendisk.sizes + part1, 0, npart * sizeof(int));
	memset(sm_gendisk.part + part1, 0, npart * sizeof(struct hd_struct));
	/* Set number of devices in struct gendisk */
	sm_gendisk.nr_real = 1;		/* FIXME for >1 device support */
	/* Do read the table */
	printk(KERN_INFO "partition check:");
	resetup_one_dev(&sm_gendisk, devn);

	/* Seems that the only way to get the results of partition scan is the following */
	for (i = 0; i < npart && sm_gendisk.part[part1 + i].start_sect > 0; i++);
	dev->nparts = i;

	return 0;
}


static int sm_revalidate(kdev_t kdev)
{
	struct sm_device *dev = &(sm_devices[DEVICE_NR(kdev)]);
	int res;

	spin_lock(&dev->lock);
	res = sm_do_revalidate(dev);
	spin_unlock(&dev->lock);
	return res;
}

static int sm_do_revalidate(struct sm_device *dev)
{
	unsigned short id, tmp;
	int i;
	
	/* Forget old information (if any) */
	dev->params = 0;
	dev->vendor = 0;
	if (dev->p2l) {
		kfree(dev->p2l);
		dev->p2l = 0;
	}
	if (dev->l2p) {
		kfree(dev->l2p);
		dev->l2p = 0;
	}
	dev->nparts = 0;
	dev->nmapped = dev->nfree = dev->nbad = dev->ncis = 0;
	sm_gendisk.nr_real = 0;			/* FIXME for >1 device support */
	/* dev->nopen and dev->npartopen should NOT be cleared here! */
	for (i = 0; i < dev->dirty_buffer_count; i++)
		kfree(dev->dirty_buffers[i]);
	dev->dirty_buffer_count = 0;
	
	/* Get identification information */
	id = sm_read_id(dev);

	/* Detect card params */
	for (dev->params = &(known_cards[0]); dev->params->id; dev->params++)
		if (dev->params->id == (id & 0xFF))
			break;
	if (!dev->params->id) {
		printk(KERN_ERR DEVICE_NAME_F ": card is either absent or not supported (id=%04x)\n",
				DEVICE_NAME_L(dev), id);
		return -EIO;
	}

	/* Detect card vendor */
	for (dev->vendor = &(known_vendors[0]); dev->vendor->id; dev->vendor++)
		if (dev->vendor->id == (id >> 8))
			break;
	if (!dev->vendor->id) {
		printk(KERN_WARNING DEVICE_NAME_F ": card vendor unknown\n", DEVICE_NAME_L(dev));
		/* unknown vendor is not fatal */
	}

	/* Print what was detected */
	printk(KERN_INFO DEVICE_NAME_F ": detected %s %s\n",
			DEVICE_NAME_L(dev), dev->vendor->name, dev->params->name);

	/* FIXME: maybe check for valid CIS header should be added */
	/* FIXME: maybe both copies of block address field should be read and compared */

	/* Build block mapping tables */
	dev->p2l = (unsigned short *) kmalloc(dev->params->pblocks * sizeof(unsigned short), GFP_KERNEL);
	if (!dev->p2l) {
		printk(KERN_ERR DEVICE_NAME_F ": failed to allocate memory for mapping table\n",
				DEVICE_NAME_L(dev));
		return -ENOMEM;
	}
	for (i = 0; i < dev->params->pblocks; i++) {
		char spare[SM_SPARE_SIZE];
		sm_read_page(dev, i, 0, 0, spare);

		/* Check for bad block */
		if (sm_is_bad_block_spare(spare)) {
			dev->p2l[i] = BAD_PBLOCK;
			dev->nbad++;
			continue;
		}

		/* Decode block mapping */
		tmp = sm_decode_baddr(sm_get_baddr_from_spare(spare), i);
		if (tmp == BAD_PBLOCK)	/* try to use copy if things go wrong */
			tmp = sm_decode_baddr(sm_get_baddr_copy_from_spare(spare), i);

		/* Update statistics */
		if (tmp == FREE_PBLOCK)
			dev->nfree++;
		else if (tmp == CIS_PBLOCK)
			dev->ncis++;
		else if (tmp == BAD_PBLOCK) {
			printk(KERN_WARNING DEVICE_NAME_F
				": failed to decode block address data of pblock %d, assuming pblock is bad\n",
				DEVICE_NAME_L(dev), i);
			dev->nbad++;
		} else
			dev->nmapped++;

		dev->p2l[i] = tmp;
	}
	printk(KERN_INFO DEVICE_NAME_F ": %u pblocks: %u mapped, %u bad, %u cis, %u free, %u kb in pblock\n",
			DEVICE_NAME_L(dev), dev->params->pblocks,
			dev->nmapped, dev->nbad, dev->ncis, dev->nfree,
			dev->params->pagesinblock * SM_PAGE_SIZE / 1024);

	dev->l2p = (unsigned short *) kmalloc(dev->params->lblocks * sizeof(unsigned short), GFP_KERNEL);
	if (!dev->l2p) {
		printk(KERN_ERR DEVICE_NAME_F ": failed to allocate memory for mapping table\n",
				DEVICE_NAME_L(dev));
		kfree(dev->p2l);
		dev->p2l = 0;
		return -ENOMEM;
	}
	/* Fast write NO_PBLOCK everywhere ... */
	memset(dev->l2p, 0xFF, dev->params->lblocks * sizeof(unsigned short));
	for (i = 0; i < dev->params->pblocks; i++) {
		if ((tmp = dev->p2l[i]) < dev->params->lblocks) {
			if (dev->l2p[tmp] != NO_PBLOCK) {
				printk(KERN_WARNING DEVICE_NAME_F
					": lblock %u is mapped both to pblock %u and pblock %u, ignoring second mapping\n",
					DEVICE_NAME_L(dev), tmp, dev->l2p[tmp], i);
			} else
				dev->l2p[tmp] = i;
		}
	}

	/* Prepare dirty buffers */
	for (i = 0; i < MAX_DIRTY_BUFFERS; i++) {
		dev->dirty_buffers[i] = kmalloc(sizeof(struct dirty_buffer)
				+ dev->params->pagesinblock * sizeof(struct dirty_buffer_page), GFP_KERNEL);
		if (!dev->dirty_buffers[i])
			break;
		dev->dirty_buffers[i]->lblock = UNUSED_BUFFER;
	}
	if (i == 0) {
		printk(KERN_ERR DEVICE_NAME_F ": failed to allocate memory for even one dirty buffer\n",
				DEVICE_NAME_L(dev));
		kfree(dev->l2p);
		dev->l2p = 0;
		kfree(dev->p2l);
		dev->p2l = 0;
		return -ENOMEM;
	}
	dev->dirty_buffer_count = i;

	/* Set block device size and read partition information */
	sm_read_part_table(dev);
	
	return 0;
}


static int sm_open(struct inode *inode, struct file *file)
{
	kdev_t kdev = inode->i_rdev;
	unsigned char devn = DEVICE_NR(kdev);
	unsigned char devminor = devn << SM_SHIFT;
	struct sm_device *dev;
	
	/* Don't allow to open nonexisting device */
	if (devn >= SM_MAXDEVS)
		return -ENODEV;

	dev = &(sm_devices[devn]);
	spin_lock(&dev->lock);
	if (dev->nopen == 0) {
		check_disk_change(kdev);
		if (!dev->p2l) {
			spin_unlock(&dev->lock);
			return -EIO;	/* no card or other problem */
		}
	} else if (dev->rawmode) {	/* don't allow second open if device is in raw mode */
		spin_unlock(&dev->lock);
		return -EBUSY;
	}

	/* Now partition count is available; don't allow to open nonexisting partition */
	if ((MINOR(kdev) & ((1 << SM_SHIFT) - 1)) > dev->nparts) {
		spin_unlock(&dev->lock);
		return -ENODEV;
	}

	/* If device was not opened yet, start the flush timer */
	if (dev->nopen == 0) {
		dev->timer.function = (void (*)(unsigned long))sm_timeout;
		dev->timer.data = (unsigned long)dev;
		dev->timer.expires = jiffies + HZ * DIRTY_FLUSH_PERIOD;
		add_timer(&dev->timer);
	}

	dev->nopen++;
	if (devminor != MINOR(kdev))
		dev->npartopen++;
	/*MOD_INC_USE_COUNT;  needed only for module? */
	spin_unlock(&dev->lock);
	return 0;
}


static int sm_close(struct inode *inode, struct file *file)
{
	kdev_t kdev = inode->i_rdev;
	unsigned char devn = DEVICE_NR(kdev);
	unsigned char devminor = devn << SM_SHIFT;
	struct sm_device *dev = &(sm_devices[devn]);
	unsigned long flags;
	
	sync_dev(kdev);
	spin_lock(&dev->lock);
	flush_dirty_buffers(dev, 1);

	spin_lock_irqsave(&dev->timer_lock, flags);	/* needed to avoid race in timer deletion */
	dev->nopen--;
	if (devminor != MINOR(kdev))
		dev->npartopen--;
	if (dev->nopen == 0)
		del_timer(&dev->timer);
	spin_unlock_irqrestore(&dev->timer_lock, flags);

	if (dev->nopen == 0 && dev->rawmode) {
		dev->rawmode = 0;
		sm_do_revalidate(dev);
	}

	/*MOD_DEC_USE_COUNT;  needed only for module? */
	spin_unlock(&dev->lock);
	return 0;
}


static int sm_ioctl(struct inode *inode, struct file *file, unsigned cmd, unsigned long arg)
{
	kdev_t kdev = inode->i_rdev;
	unsigned char devn = DEVICE_NR(kdev);
	unsigned char devminor = devn << SM_SHIFT;
	struct sm_device *dev = &(sm_devices[devn]);
	struct hd_geometry geo;
	struct sm_card_info *ci;
	struct sm_card_stats *cs;
	struct sm_page_request *pr;
	short *tab;
	int res;

	switch (cmd) {
		
		case SM_GET_CARD_INFO:
			if (arg == 0)
				return -EINVAL;
			ci = (struct sm_card_info *) arg;
			if (!access_ok(VERIFY_WRITE, ci, sizeof(struct sm_card_info)))
				return -EFAULT;
			if (devminor != MINOR(kdev))
				return -ENOTTY;		/* allow only for whole device */
			if (!(file->f_mode & FMODE_READ))
				return -EPERM;
			spin_lock(&dev->lock);
			ci->pblocks = dev->params->pblocks;
			ci->lblocks = dev->params->lblocks;
			ci->pagesinblock = dev->params->pagesinblock;
			spin_unlock(%dev->lock);
			return 0;
			
		case SM_GET_P2L_MAP:
			if (arg == 0)
				return -EINVAL;
			tab = (unsigned short *) arg;
			if (devminor != MINOR(kdev))
				return -ENOTTY;		/* allow only for whole device */
			spin_lock(&dev->lock);
			if (!access_ok(VERIFY_WRITE, tab, dev->params->pblocks * sizeof(unsigned short))) {
				spin_unlock(&dev->lock);
				return -EFAULT;
			}
			if (dev->rawmode) {
				spin_unlock(&dev->lock);
				return -EBUSY;
			}
			memcpy(tab, dev->p2l, dev->params->pblocks * sizeof(unsigned short));
			spin_unlock(%dev->lock);
			return 0;

		case SM_GET_L2P_MAP:
			if (arg == 0)
				return -EINVAL;
			tab = (unsigned short *) arg;
			if (devminor != MINOR(kdev))
				return -ENOTTY;		/* allow only for whole device */
			spin_lock(&dev->lock);
			if (!access_ok(VERIFY_WRITE, tab, dev->params->lblocks * sizeof(unsigned short))) {
				spin_unlock(&dev->lock);
				return -EFAULT;
			}
			if (dev->rawmode) {
				spin_unlock(&dev->lock);
				return -EBUSY;
			}
			memcpy(tab, dev->l2p, dev->params->lblocks * sizeof(unsigned short));
			spin_unlock(%dev->lock);
			return 0;

		case SM_PHYS_READ_RAW:
		case SM_PHYS_READ:
			if (!(file->f_mode & FMODE_READ))
				return -EACCES;
			if (arg == 0)
				return -EINVAL;
			pr = (struct sm_page_request *) arg;
			if (!access_ok(VERIFY_READ, pr, sizeof(struct sm_page_request)))
				return -EFAULT;
			if (devminor != MINOR(kdev))
				return -ENOTTY;		/* allow only for whole device */

			spin_lock(&dev->lock);
			if (pr->pblock >= dev->params->pblocks || pr->page >= dev->params->pagesinblock) {
				spin_unlock(&dev->lock);
				return -EINVAL;
			}
			if (cmd == SM_PHYS_READ_RAW) {
				sm_read_page(dev, pr->pblock, pr->page, pr->data, pr->spare);
				res = 0;
			} else
				res = sm_read_and_correct_page(dev, pr->pblock, pr->page, pr->data, pr->spare);
			spin_unlock(&dev->lock);
			return res;

		case SM_PHYS_WRITE_RAW:
		case SM_PHYS_WRITE:
			if (!capable(CAP_SYS_ADMIN))
				return -EACCES;
			if (arg == 0)
				return -EINVAL;
			pr = (struct sm_page_request *) arg;
			if (!access_ok(VERIFY_WRITE, pr, sizeof(struct sm_page_request)))
				return -EFAULT;
			if (devminor != MINOR(kdev))
				return -ENOTTY;		/* allow only for whole device */

			spin_lock(&dev->lock);
			if (pr->pblock >= dev->params->pblocks || pr->page >= dev->params->pagesinblock) {
				spin_unlock(&dev->lock);
				return -EINVAL;
			}
			if (dev->nopen > 1) {
				spin_unlock(&dev->lock);
				return -EBUSY;
			}
			if (!dev->rawmode) {
				fsync_dev(kdev);
				invalidate_buffers(kdev);
				dev->rawmode = 1;
			}
			if (cmd == SM_PHYS_WRITE) {
				sm_calculate_ecc(pr->data, sm_spare_ecc1(pr->spare));
				sm_calculate_ecc(pr->data + 256, sm_spare_ecc2(pr->spare));
				sm_put_baddr_to_spare(sm_get_baddr_from_spare(pr->spare), pr->spare);
			}
			res = sm_write_page(dev, pr->pblock, pr->page, pr->data, pr->spare);
			spin_unlock(&dev->lock);
			return res;

		case SM_PHYS_ERASE_BLOCK:
		case SM_PHYS_MARK_BAD_BLOCK:
			if (!capable(CAP_SYS_ADMIN))
				return -EACCES;
			spin_lock(&dev->lock);
			if (arg >= dev->params->pblocks) {
				spin_unlock(&dev->lock);
				return -EINVAL;
			}
			if (dev->nopen > 1) {
				spin_unlock(&dev->lock);
				return -EBUSY;
			}
			if (!dev->rawmode) {
				fsync_dev(kdev);
				invalidate_buffers(kdev);
				dev->rawmode = 1;
			}
			if (cmd == SM_PHYS_ERASE_BLOCK)
				res = sm_erase_block(dev, arg);
			else {
				mark_bad_pblock(dev, arg);
				res = 0;
			}
			spin_unlock(&dev->lock);
			return res;
		
		case SM_GET_CARD_STATS:
			if (arg == 0)
				return -EINVAL;
			cs = (struct sm_card_stats *) arg;
			if (!access_ok(VERIFY_WRITE, cs, sizeof(struct sm_card_stats)))
				return -EFAULT;
			if (devminor != MINOR(kdev))
				return -ENOTTY;		/* allow only for whole device */
			spin_lock(&dev->lock);
			if (dev->rawmode) {
				spin_unlock(&dev->lock);
				return -EBUSY;
			}
			cs->nmapped = dev->nmapped;
			cs->nbad = dev->nbad;
			cs->ncis = dev->ncis;
			cs->nfree = dev->nfree;
			spin_unlock(%dev->lock);
			return 0;

		/* Information about emulated hard drive */

		case HDIO_GETGEO:
			spin_lock(&dev->lock);
			geo.cylinders = dev->params->cylinders;
			geo.heads = dev->params->heads;
			geo.sectors = dev->params->sectors;
			geo.start = sm_gendisk.part[devminor].start_sect;
			spin_unlock(&dev->lock);
			return copy_to_user((void *)arg, &geo, sizeof(geo));

		/* Partition table rescan */

		case BLKRRPART:
			if (!capable(CAP_SYS_ADMIN))
				return -EACCES;
			spin_lock(&dev->lock);
			if (dev->npartopen > 0) {
				spin_unlock(&dev->lock);
				return -EBUSY;
			}
			res = sm_read_part_table(dev);
			spin_unlock(&lock);
			return res;

		/* Standard block device ioctls */

		case BLKGETSIZE:
			if (!arg)
				return -EINVAL;
			return put_user(sm_devsizes[MINOR(kdev)] * (SM_BLOCKSIZE / SM_PAGE_SIZE), (long *)arg);

		case BLKFLSBUF:
			if (!capable(CAP_SYS_ADMIN))
				return -EACCES;
			fsync_dev(kdev);
			invalidate_buffers(kdev);
		        return 0;

		case BLKRAGET:
			if (!arg)
				return -EINVAL;
			return put_user(read_ahead[MAJOR(kdev)], (long *)arg);
			
		case BLKRASET:
			if (!capable(CAP_SYS_ADMIN))
				return -EACCES;
			if (arg > 0x80)
				return -EINVAL;
			read_ahead[MAJOR(kdev)] = arg;
			return 0;

		RO_IOCTLS(kdev, arg);
	}

	return -ENOTTY;
}

static int sm_fsync(struct file *file, struct dentry *dentry)
{
	int i, res, res2;

	/* First call default fsync() for block device */
	res = block_fsync(file, dentry);

	/* Now flush all dirty buffers in all devices */
	for (i = 0; i < SM_MAXDEVS; i++) {
		struct sm_device *dev = &(sm_devices[i]);
		spin_lock(&dev->lock);
		res2 = flush_dirty_buffers(dev, 1);
		if (!res && res2)
			res = res2;
		spin_lock(&dev->lock);
	}

	return res;
}
