#ifndef LINUX_SM_H
#define LINUX_SM_H

#ifndef __KERNEL__
#include <string.h>
#endif

/* Sizes of page and spare area */
#define SM_PAGE_SIZE 512
#define SM_SPARE_SIZE 16

/* Spare area offsets */
#define SM_SAOFFS_BADBLOCK	5	/* bad block mark */
#define SM_SAOFFS_BLOCKADDR_H	6	/* 8..15 bits of block address data */
#define SM_SAOFFS_BLOCKADDR_L	7	/* 0..7 bits of block address data */
#define SM_SAOFFS_ECC2		8	/* ECC of second half of the page */
#define SM_SAOFFS_BACOPY_H	11	/* 8..15 bits if block address data copy */
#define SM_SAOFFS_BACOPY_L	12	/* 0..7 bits of block address data copy */
#define SM_SAOFFS_ECC1		13	/* ECC of first half of the page */

/* Value written to spare area at SM_SAOFFS_BADBLOCK offset to mark bad block */
#define SM_SA_BADBLOCK_VALUE	0xF0

/* This many bytes at the beginning of spare area should be 0xFF's for good blocks */
#define SM_SA_UNUSEDFORGOOD	6

/* Number of physical blocks in every but last zone */
#define PBLOCKSINZONE	1024
/* Number of logical blocks in every but last zone */
#define LBLOCKSINZONE	1000

/* Get zone number for physical block */
#define PZONE(pblock) (pblock / PBLOCKSINZONE)
/* Get zone number for logical block */
#define LZONE(lblock) (lblock / LBLOCKSINZONE)

/* Structs used in ioctl's */
struct sm_card_info {
	unsigned short pblocks;		/* total number of physical blocks */
	unsigned short lblocks;		/* total number of logical blocks */
	unsigned short pagesinblock;	/* number of pages in each physical block */
};
struct sm_page_request {
	unsigned short pblock;		/* number of physical block */
	unsigned short page;		/* number of page in the block */
	char data[SM_PAGE_SIZE];	/* data buffer */	
	char spare[SM_SPARE_SIZE];	/* spare area buffer */
};
struct sm_card_stats {
	unsigned short nmapped;	/* number of mapped blocks */
	unsigned short nbad;	/* number of bad blocks */
	unsigned short ncis;	/* number of cis blocks (should be 1 for valid cards) */
	unsigned short nfree;	/* number of free blocks */
};

/* Special values for pblock to lblock mappings */
#define FREE_PBLOCK	0xFFFF		/* free block */
#define BAD_PBLOCK	0xFFFE		/* bad block */
#define CIS_PBLOCK	0xFFFD		/* CIS block */

/* Special value for lblock to pblock mapping */
#define NO_PBLOCK	0xFFFF		/* unallocated block */

/* ioctl codes */

/* Get card information. Third parameter should be (struct sm_card_info *) */
#define SM_GET_CARD_INFO	0x3890

/* Get physical-to-logical block mapping information.
 * Third parameter should be a pointer to an array of unsigned shorts.
 * This ioctl is not available if device is in raw mode */
#define SM_GET_P2L_MAP		0x3891

/* Get logical-to-physical block mapping information.
 * Third parameter should be a pointer to an array of unsigned shorts
 * This ioctl is not available if device is in raw mode */
#define SM_GET_L2P_MAP		0x3892

/* Read physical page and spare area. No ECC correction performed.
 * Third parameter should be (struct sm_page_request *) */
#define SM_PHYS_READ_RAW	0x3893

/* Read physical page and spare area, and correct data using ECC. -EIO is returned if correction fails.
 * Third parameter should be (struct sm_page_request *) */
#define SM_PHYS_READ		0x3894

/* Write physical page and spare area. No checks on data are performed.
 * This ioctl first turns device into raw mode (and fails if this is not possible).
 * Third parameter should be (struct sm_page_request *) */
#define SM_PHYS_WRITE_RAW	0x3895

/* Write physical page.
 * This ioctl first turns device into raw mode (and fails if this is not possible).
 * Third parameter should be (struct sm_page_request *)
 * In spare area, ecc and block address data copy are filled automatically (and returned). */
#define SM_PHYS_WRITE		0x3896

/* Erase physical block.
 * This ioctl first turns device into raw mode (and fails if this is not possible).
 * Third parameter should be block number */
#define SM_PHYS_ERASE_BLOCK	0x3897

/* Mark physical block as bad (that is, try to erase and then write bad block marks
 * to spare area of all pages).
 * This ioctl first turns device into raw mode (and fails if this is not possible).
 * Third parameter should be block number */
#define SM_PHYS_MARK_BAD_BLOCK	0x3898

/* Get statustics. Third parameter should be (struct sm_card_stats *)
 * This ioctl is not available in raw mode */
#define SM_GET_CARD_STATS	0x3899


/* Count number of ones in bit representation of param */
static inline int ones(unsigned long param)
{
	int n = 0;
	while (param) {
		n += (param & 1);
		param >>= 1;
	}
	return n;
}

/* Build the beginning of spare area for good block */
static inline void sm_build_good_block_spare_start(unsigned char *spare)
{
	memset(spare, 0xFF, SM_SA_UNUSEDFORGOOD);
}

/* Build spare area for bad block */
static inline void sm_build_bad_block_spare(unsigned char *spare)
{
	memset(spare, 0xFF, SM_SPARE_SIZE);
	spare[SM_SAOFFS_BADBLOCK] = SM_SA_BADBLOCK_VALUE;
}

/* Check spare area for bad block mark */
static inline int sm_is_bad_block_spare(unsigned char *spare)
{
	return ones(spare[SM_SAOFFS_BADBLOCK]) <= 6;
}

/* Build block address data field for given lblock */
static inline unsigned short sm_encode_baddr(unsigned short lblock)
{
	unsigned short baddr;

	if (lblock == CIS_PBLOCK)
		return 0;
	if (lblock == FREE_PBLOCK)
		return 0xFFFF;

	baddr = ((lblock % LBLOCKSINZONE) << 1) | 0x1000;
	baddr |= (ones(baddr) & 0x1);
	return baddr;
}

/* Get block address data from spare area */
static inline unsigned short sm_decode_baddr(unsigned short baddr, unsigned short pblock)
{
	unsigned short tmp;
	
	if (baddr == 0)
		return CIS_PBLOCK;
	if (baddr == 0xFFFF)
		return FREE_PBLOCK;
	if (ones(baddr) & 0x1)
		return BAD_PBLOCK;
	tmp = (baddr >> 1) & 0x3FF;
	if (tmp >= LBLOCKSINZONE)
		return BAD_PBLOCK;
	return PZONE(pblock) * LBLOCKSINZONE + tmp;
}

/* Put encoded block address data field into spare area */
static inline void sm_put_baddr_to_spare(unsigned short baddr, unsigned char *spare)
{
	spare[SM_SAOFFS_BLOCKADDR_H] = spare[SM_SAOFFS_BACOPY_H] = (baddr >> 8);
	spare[SM_SAOFFS_BLOCKADDR_L] = spare[SM_SAOFFS_BACOPY_L] = (baddr & 0xFF);
}

/* Get encoded block address data field from spare area */
static inline unsigned short sm_get_baddr_from_spare(unsigned char *spare)
{
	return (spare[SM_SAOFFS_BLOCKADDR_H] << 8) | spare[SM_SAOFFS_BLOCKADDR_L];
}

/* Get encoded block address data copy from spare area */
static inline unsigned short sm_get_baddr_copy_from_spare(unsigned char *spare)
{
	return (spare[SM_SAOFFS_BACOPY_H] << 8) | spare[SM_SAOFFS_BACOPY_L];
}

/* Get ECC1 field address in spare area */
static inline unsigned char *sm_spare_ecc1(unsigned char *spare)
{
	return spare + SM_SAOFFS_ECC1;
}

/* Get ECC2 field address in spare area */
static inline unsigned char *sm_spare_ecc2(unsigned char *spare)
{
	return spare + SM_SAOFFS_ECC2;
}

#endif
