#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>

#define DEFAULT_PORT 3549
#define CHUNKSIZE 528		/* 512 data + 16 spare */

int doread(int fd, char *buf, int count)
{
	int got = 0, res;
	while (count > 0) {
		res = read(fd, buf, count);
		if (res < 0)
			return res;
		if (res == 0)
			return got;
		got += res;
		buf += res;
		count -= res;
	}
	return got;
}

int dowrite(int fd, char *buf, int count)
{
	int got = 0, res;
	while (count > 0) {
		res = write(fd, buf, count);
		if (res < 0)
			return res;
		if (res == 0)
			return got;
		got += res;
		buf += res;
		count -= res;
	}
	return got;
}


int main(int argc, char *argv[])
{
	enum { NoCmd, BackupCmd, RestoreCmd } cmd = NoCmd;
	char *port = 0, *filename = 0, *p;
	struct sockaddr_in sa;
	static struct option options[] = {
		{ "port", required_argument, 0, 'p' },
		{ "backup", no_argument, 0, 'b' },
		{ "restore", no_argument, 0, 'r' },
		{ 0, 0, 0, 0 }
	};
	int fd, sock, s;
	int c;
	char buf[CHUNKSIZE];
	
	opterr = 0;	/* shut up getopt()'s internal diagnostics */

	if (argc == 1)
		goto usage;

	while (1) {
		c = getopt_long(argc, argv, "p:br", options, 0);
		if (c == -1)
			break;
		switch (c) {
			case 'p':
				if (port) {
					fprintf(stderr, "smhelper: duplicate port specification\n");
					return 1;
				}
				port = optarg;
				for (p = port; *p >= '0' && *p <= '9'; p++);
				if (*p || p == port) {
					fprintf(stderr, "smhelper: invalid port specification\n");
					return 1;
				}
				break;
			case 'b':
				if (cmd != NoCmd)
					goto usage;
				cmd = BackupCmd;
				break;
			case 'r':
				if (cmd != NoCmd)
					goto usage;
				cmd = RestoreCmd;
				break;
		}
	}

	if (optind != argc - 1)
		goto usage;
	filename = argv[optind];

	if (cmd == NoCmd)
		goto usage;

	if (cmd == BackupCmd)
		fd = open(filename, O_WRONLY | O_CREAT, 0644);
	else
		fd = open(filename, O_RDONLY);
	if (fd < 0) {
		fprintf(stderr, "smhelper: open %s: %s\n", filename, strerror(errno));
		return 1;
	}
		
	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = htonl(INADDR_ANY);
	sa.sin_port = htons(port ? atoi(port) : DEFAULT_PORT);

	sock = socket(PF_INET, SOCK_STREAM, 0);
	if (sock < 0) {
		perror("smhelper: socket");
		return 1;
	}

	if (bind(sock, (struct sockaddr *)(&sa), sizeof(sa)) < 0) {
		perror("smhelper: bind");
		return 1;
	}
	if (listen(sock, 1) < 0) {
		perror("smhelper: listen");
		return 1;
	}
	s = accept(sock, 0, 0);
	if (s < 0) {
		perror("smhelper: accept");
		return 1;
	}
	
	if (cmd == BackupCmd) {
		printf("smhelper: connected, saving backup to %s\n", filename);
		while (1) {
			c = doread(s, buf, CHUNKSIZE);
			if (c == 0)
				break;
			if (c < 0) {
				perror("smhelper: read from socket");
				return 1;
			}
			if (c < CHUNKSIZE) {
				fprintf(stderr, "smhelper: read from socket: short read\n");
				return 1;
			}
			c = dowrite(fd, buf, CHUNKSIZE);
			if (c < 0) {
				perror("smhelper: write");
				return 1;
			}
			if (c < CHUNKSIZE) {
				fprintf(stderr, "smhelper: short write\n");
				return 1;
			}
		}
	} else {
		printf("smhelper: connected, restoring backup from %s\n", filename);
		while (1) {
			c = doread(fd, buf, CHUNKSIZE);
			if (c == 0)
				break;
			if (c < 0) {
				perror("smhelper: read");
				return 1;
			}
			if (c < CHUNKSIZE) {
				fprintf(stderr, "smhelper: short read\n");
				return 1;
			}
			c = dowrite(s, buf, CHUNKSIZE);
			if (c < 0) {
				perror("smhelper: write to socket");
				return 1;
			}
			if (c < CHUNKSIZE) {
				fprintf(stderr, "smhelper: write to socket: short write\n");
				return 1;
			}
		}
	}

	close(s);
	close(sock);
	close(fd);
	return 0;

usage:
	fprintf(stderr, "Usage:\n"
			"  smhelper [--port port] --backup filename\n"
			"    backup data received from tcp socket to filename.\n"
			"  smhelper [--port port] --restore filename\n"
			"    send data from file to tcp_socket.\n"
			"  --port port\n"
			"    wait for connection on altername port instead of %d\n", DEFAULT_PORT);
	return 1;
}
